// A Rule is not really an Element (since it can not occur as part of a
// grammar rule), but it otherwise shares a lot of the features.
// Tue Nov 10 14:57:16 1998, tiggr@gerbil.org
implementation class
Rule: Element

end;

implementation instance
Rule
{
  <doc> The name of the non-terminal.  </doc>
  public String name;

  <doc> The expression to be evaluated before the first grammar code.  </doc>
  String expr;

  <doc> The expression to be evaluated after the end of the last grammar
      code.  </doc>
  public mutable String post_expr;

  <doc> The productions.  </doc>
  Alternatives rhs;

  <doc> The attributes.  </doc>
  public Array attributes;

  <doc> Whether we're the top rule.  </doc>
  mutable boolean is_top;

  <doc> Iff TRUE, we've checked for LL1'ness.  </doc>
  boolean checked_ll1;
}

void
  checkLL1
{
  if (!checked_ll1)
    {
      checked_ll1 = TRUE;
      [rhs checkLL1];
    }
}

void
  makeFlatRules
{
  [rhs makeFlatRules name];
}

<doc> Print a tuple of the strings returned by asking the attributes
    that match {in_not_out} their {desired_string}.  Return the number
    of attributes that matched.  Do not print anything when nothing
    matched.  </doc>
int (num_done)
	   print OutputStream s
  attributeTuple boolean in_not_out
	   using selector desired_string
{
  int i, n_attr = [attributes length];
  for (i = 0; i < n_attr; i++)
    {
      Attribute a = attributes[i];

      if ([a in_not_out] == in_not_out)
	{
	  if (!num_done++)
	    [s print '('];
	  else
	    [s print ", "];

	  [s print ({String s = [a perform desired_string];})];
	}
    }
  if (num_done > 0)
    [s print ')'];
}

OutputStream
  compile OutputStream s
{
  int i, n_attr, n_attr_out;
  Attribute a;

  [s print indent];

  /* Output the return type, aka the type of the synthesized attributes.  */
  if (attributes != nil)
    {
      n_attr = [attributes length];
      n_attr_out = [self print s attributeTuple FALSE
			 using selector (String type)];
      if (n_attr_out != 0)
	[self print s attributeTuple FALSE using selector (String name)];
    }

  if (!n_attr_out)
    [s print "void"];
  [s print (indent, "  ", name)];

  /* Output the arguments, aka the inherited attributes.  */
  if (n_attr != 0 && n_attr != n_attr_out)
    {
      [self print s attributeTuple TRUE using selector (String type)];
      [self print s attributeTuple TRUE using selector (String name)];
    }

  /* Start the body of the method.  */
  [s print (indent, "{")];
  [indent add ' '];

  if (is_top)
    {
      [s print (indent, "gps.Parser previous_parser = current_parser;")];
      [s print (indent, "current_parser = self;")];
      [s print (indent, "unwind (current_parser = previous_parser)")];
      [s print (indent, "({")];
      [indent add ' '];
      [s print (indent, "[self nextToken];")];
    }

  [s print (indent, "if (trace) [[err print \"entering rule: ",
	    name, "\"] nl];")];

  if (expr != nil)
    [s print (indent, expr)];

  [rhs compile s];

  if (is_top)
    {
      [s print (indent, "if (token != EOF) [self fatalError];")];
      [indent removeElementAt 1];
      [s print (indent, "});")];
    }

  if (post_expr != nil)
    [s print (indent, post_expr)];

  [indent removeElementAt 1];
  [s print (indent, "}")];

  = s;
}

IntegerRangeSet
  computeLA IntegerRangeSet f
{
  if (!follow)
    follow = [IntegerRangeSet new];
  else if (run == current_run && [f equal follow])
    return la;
  run = current_run;

  [follow uniteWith f];

  la = [rhs computeLA follow];

  if (first == nil || ![first equal [rhs first]])
    {
      changes = YES;
      first = [rhs first];
    }

  = la;
}

id
	 init String n
	start String e
  productions Alternatives p
   attributes Array a
{
  (name, expr, rhs, attributes) = (n, e, p, a);

  = [super init];
}

end;
