<copyright> GP Lexer interface and skeleton.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996-1998 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Lexer.t,v 1.13 1998/09/15 09:06:13 tiggr Exp $</id>
    </copyright>

implementation class
Lexer: instance (All)

end;

implementation instance
Lexer: instance (All)

<doc> Return the current line at which this lexer is reading.  </doc>
deferred int
  current_line;

<doc> Return the name of the file being read.  </doc>
deferred String
  filename;

<doc> Return the Parser (currently) being fed by this lexer.  </doc>
deferred Parser
  parser;

<doc> Read and return another token from the input.  </doc>
deferred int
  nextToken;

<doc> Return the semantic meaning of the current token, in any way the
    sender wants.  </doc>
deferred Any
  semantics;

<doc> Set the current parser.  This is the parser that will be returned by
    the {parser} method.  </doc>
deferred void
  set_parser Parser parser;

<doc> Return the current token.  </doc>
deferred int
  token;

end;

/******************** Skeleton ********************/

implementation class
SkeletonLexer: Lexer, State

end;

implementation instance
SkeletonLexer
{
  <doc> The parser which we're feeding.  </doc>
  public Parser parser;

  <doc> The current line number.  </doc>
  public int current_line;

  <doc> The filename, at least what we think it is.  </doc>
  public String filename;

  <doc> The current token.  </doc>
  public int token;

  <doc> The semantic content of the current token.  </doc>
  All semantics;
}

<doc> Designated initializer.  </doc>
id
  initWithFilename String name
	     line: int start_line = 1
{
  (current_line, filename) = (start_line, name);
  (parser, semantics) = (nil, nil);

  = self;
}

<doc> Return the {semantics} any way the sender wants.  </doc>
Any
  semantics
{
  = Any (semantics);
}

void
  set_parser Parser p
pre
  !parser
{
  parser = p;
}

end;
