<copyright> GP Parser skeleton.
    Written by <a href="mailto:tiggr@ics.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Parser.t,v 1.24 1998/05/18 09:35:58 tiggr Exp $</id>
    </copyright>

/******************** ParserDefines ********************/

implementation class
ParserDefines
{
  // This should not be 0, obviously.  However, 0 is what lex returns for
  // end of file...
  // Sun Jun  9 16:27:08 1996, tiggr@cobra.es.ele.tue.nl
  const EOF = 0;
}

end;

implementation instance ParserDefines end;

/******************** FileElement ********************/

implementation class
FileElement: instance (All)
end;

implementation instance
FileElement: instance (All)

<doc> Return the line in the file at which this element was declared (or
    whatever you call it).  </doc>
deferred int
  line;

<doc> Describe the location of the receiving file element to the stream
    {s}.  The default implementation prints the line.  The {parser} may be
    {nil}, if there is no current parser.  </doc>
OutputStream
  describeLocation OutputStream s
	       for Parser parser
{
  if (parser != nil)
    s = [s print ([[parser lex] filename], ':')];
  = [s print [self line]];
}

end;

/******************** Parser ********************/

implementation class
Parser: State, stdio
{
  <doc> The path for all parsers to search.  </doc>
  static MutableIndexed path;

  <doc> The current parser.  </doc>
  static public Parser current_parser;

  <doc> The global number of errors counted.  </doc>
  int global_num_errors;

  <doc> The global number of warnings counted.  </doc>
  int global_num_warnings;

  <doc> Mapping from token number to token name for this parser class.
      </doc>
  MutableArray tokens_by_number;

  <doc> A default value for tracing, per parser class.  </doc>
  public boolean trace_default;
}

/******************** Tracing ********************/
<doc> <h4>Tracing</h4> </doc>

<doc> Output information on the {ByteString} unit arguments.  </doc>
OutputStream (s)
  help OutputStream s
  done MutableKeyed done
{
  class (State) cls = [Parser self];

  if (!done[cls])
    {
      [done add cls];
      [[s print "gps.Parser
  :trace-parser <class>	set trace_default of the parser <class>"] nl];
    }
}

<doc> Digest any {:trace-parser} arguments.  </doc>
void
  load MutableArray arguments
{
  int i, n = [arguments length];

  while (i < n - 1)
    if ([arguments[i] equal ":trace-parser"])
      {
	String name = arguments[i + 1];
	[arguments removeElements (i, 2)];
	n -= 2;

	class (Parser) pc = class (Parser) ([Runtime classNamed name]);
	if (!pc)
	  [[[stdio err] print ("no class named `", name, "'")] nl];
	else
	  [pc set_trace_default YES];
      }
    else
      i++;
}

void
  set_trace_default boolean value
{
  if (!tokens_by_number)
    [self fill_tokens_by_number];
  trace_default = value;
}

deferred void
  fill_tokens_by_number;

/******************** the path ********************/
<doc> <h4>The search path</h4> </doc>

<doc> Add the {directory} to the search {path}.  </doc>
void
  addToPath String directory
{
  if (!path)
    path = [MutableObjectArray new];

  [path add directory];
}

<doc> Return the {path}.  </doc>
Indexed
  path
{
  = path;
}

/******************** (error) reporting ********************/
<doc> <h4>Errors and other reports</h4> </doc>

<doc> Return the stream to which to output errors.  The default stream is
    {err}.  </doc>
OutputStream
  errorStream
{
  = [stdio err];
}

<doc> Return the {global_num_errors}.  (The variable isn't public to share
    a few selectors.)  </doc>
int
  num_errors
{
  = global_num_errors;
}

<doc> Return the {global_num_warnings}.  </doc>
int
  num_warnings
{
  = global_num_warnings;
}

<doc> Return the stream to which to report an error by asking the
    {current_parser}.  </doc>
OutputStream
  error
{
  = current_parser != nil ? [current_parser error]
      :
	({
	  global_num_errors++;
	  [[self errorStream] print ([Runtime program_name], ": ")];
	});
}

<doc> Return the stream to which to report an error for the {object},
    similar to {error}.  </doc>
OutputStream
  error FileElement object
{
  = (current_parser != nil ? [current_parser error object]
     :
       ({
	 global_num_errors++;
	 [[object describeLocation [self errorStream] for nil] print ": "];
       }));
}

<doc> Return the stream to which to report something by asking the
    {current_parser}.  </doc>
OutputStream
  report
{
  = current_parser != nil ? [current_parser report] : [self errorStream];
}

<doc> Return the stream to which to report something for the {object},
    similar to an {error} but without counting it as one.  </doc>
OutputStream
  report FileElement object
{
  = (current_parser != nil ? [current_parser report object]
     : [[object describeLocation [self errorStream] for nil] print ": "]);
}

<doc> Return the stream to which to report a warning by asking the
    {current_parser}.  </doc>
OutputStream
  warning
{
  = (current_parser != nil ? [current_parser warning]
     :
       ({
	 global_num_warnings++;
	 [[self errorStream] print ([Runtime program_name], ": warning: ")];
       }));
}

<doc> Return the stream to which to report a warning for the {object},
    similar to {warning}.  </doc>
OutputStream
  warning FileElement object
{
  = (current_parser != nil ? [current_parser warning object]
     :
       ({
	 global_num_warnings++;
	 [[object describeLocation [self errorStream] for nil]
	   print ": warning: "];
       }));
}

end;

implementation instance
Parser
{
  <doc> The lexer from which we get the tokens.  </doc>
  public Lexer lex;

  <doc> The current token.  </doc>
  int token;

  <doc> The number of errors and warnings reported through this parser.  </doc>
  public int num_errors, num_warnings;

  <doc> Iff !0, we output all kinds of nice debugging information.  </doc>
  public boolean trace;
}

<doc> Describe the {token} on the output stream {s}.  </doc>
OutputStream
  describe OutputStream s
     token int token
{
  = ({if (token >= ' ' && token <= '~')
      [s print ('\'', byte (token), '\'')];
    else if (token == 0)
      [s print "EOF"];
    else if (token >= 65536)
      if (!tokens_by_number)
	[s print token];
      else
	[tokens_by_number[token - 65536] write s];
    else
      [s print ("char(", token, ')')];});
}

<doc> Describe the tokens from the enumerator {e} on the output stream {s}.
    </doc>
OutputStream
  describe OutputStream s
    tokens Enumerator e
{
  boolean first = YES;
  int i;

  while ({boolean b; (b, i) = [e next]; b;})
    {
      if (first)
	first = NO;
      else
	s = [s print ' '];
      s = [self describe s token i];
    }
  = s;
}

<doc> Return the stream to which to output errors.  The default stream is
    {err}.  </doc>
OutputStream
  errorStream
{
  = [stdio err];
}

<doc> Return the stream to which to report an error.  </doc>
OutputStream
  error
{
  num_errors++;
  global_num_errors++;
  = [[self errorStream] print ([lex filename], ':', [lex current_line], ": ")];
}

<doc> Return the stream to which to report an error for the {object},
    similar to {error}.  </doc>
OutputStream
  error FileElement object
{
  num_errors++;
  global_num_errors++;
  = [[object describeLocation [self errorStream] for self] print ": "];
}

<doc> Return the stream to which to report something.  </doc>
OutputStream
  report
{
  = [[self errorStream] print ([lex filename], ':', [lex current_line], ": ")];
}

<doc> Return the stream to which to report something for the {object},
    similar to {error}, but without counting it as one.  </doc>
OutputStream
  report FileElement object
{
  = [[object describeLocation [self errorStream] for self] print ": "];
}

<doc> Return the stream to which to report a warning.  </doc>
OutputStream
  warning
{
  num_warnings++;
  global_num_warnings++;
  = [[self errorStream]
      print ([lex filename], ':', [lex current_line], ": warning: ")];
}

<doc> Return the stream to which to report a warning for the {object},
    similar to {warning}.  </doc>
OutputStream
  warning FileElement object
{
  num_warnings++;
  global_num_warnings++;
  = [[object describeLocation [self errorStream] for self] print ": warning: "];
}

<doc> Return the stream to which to report a syntax error.  </doc>
OutputStream
  syntaxError
{
  OutputStream s = [[self error] print "error at "];

  = [self describe s token token];
}

<doc> Issue the error and exit.  </doc>
void
  fatalError
{
  [[self syntaxError] nl];
  [Runtime exit 1];
}

/******************** misc ********************/
<doc> <h4>Miscellaneous</h4> </doc>

<doc> Designated initializer.  </doc>
id
  init Lexer l
{
  lex = l;
  trace = trace_default;

  [lex set_parser self];

  = self;
}

int
  nextToken
{
  if (trace)
    [err print "reading a token: "];

  token = [lex nextToken];

  if (trace)
    {
      [self describe err token token];

      Any semantics = [lex semantics];
      if (semantics)
	[err print (" semantics: ", semantics)];

      [err nl];
    }

  = token;
}

void
  set_trace boolean value
{
  if (value && !tokens_by_number)
    [isa fill_tokens_by_number];
  trace = value;
}

void
  term int expected
{
  if (token != expected)
    {
      OutputStream s = [self syntaxError];

      [s print (", expected token ")];
      [self describe s token expected];
      [s nl];
    }

  [self nextToken];
}

Array
  tokens_by_number
{
  = tokens_by_number;
}

end;
